# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

"""
Business logic for Coal Mine
"""

from copy import copy
import datetime
from logbook import Logger
import math
from numbers import Number
from random import choice
import re
import smtplib
import signal

log = Logger('BusinessLogic')


class AlreadyExistsError(Exception):
    pass


class AlreadyPausedError(Exception):
    pass


class AlreadyUnpausedError(Exception):
    pass


class BusinessLogic(object):
    def __init__(self, store, email_sender):
        self.store = store
        self.email_sender = email_sender

    def create(self, name, periodicity, description=None, emails=[],
               paused=False):
        canary = {'id': self.create_identifier()}

        if not isinstance(name, str):
            raise TypeError('name must be a string')
        if not name:
            raise TypeError('name must be non-empty')
        canary['name'] = name

        slug = self.slug(name)
        try:
            conflict = self.store.find_identifier(slug)
            raise AlreadyExistsError(
                'Canary {} already exists with identifier {}'.format(
                    slug, conflict))
        except KeyError:
            pass
        canary['slug'] = slug

        if not isinstance(periodicity, Number):
            raise TypeError('periodicity must be a number')
        if periodicity <= 0:
            raise TypeError('periodicity must be positive')
        canary['periodicity'] = periodicity

        if not description:
            description = ''
        if not isinstance(description, str):
            raise TypeError('description must be a string')
        canary['description'] = description

        if emails is None:  # Be generous in what you accept...
            emails = []
        if isinstance(emails, str):
            raise TypeError('emails should be a list of zero or more '
                            'addresses')
        canary['emails'] = list(emails)

        if not isinstance(paused, bool):
            raise TypeError('paused should be a bool')
        canary['paused'] = paused

        canary['history'] = [(datetime.datetime.utcnow(), 'Snitch created')]
        canary['late'] = False

        if not canary['paused']:
            canary['deadline'] = canary['history'][0][0] + \
                datetime.timedelta(seconds=periodicity)

        self.store.create(canary)

        log.info('Created canary {} ({})'.format(canary['id'],
                                                 canary_log_string(canary)))

        self.schedule_next_deadline()

        return canary

    def update(self, identifier, name=None, periodicity=None,
               description=None, emails=None):
        canary = self.store.get(identifier)
        updates = {}
        notify = False

        if name is not None and name != canary['name']:
            if not isinstance(name, str):
                raise TypeError('name must be a string')
            if not name:
                raise TypeError('name must be non-empty')
            old_slug = canary['slug']
            new_slug = self.slug(name)
            if old_slug != new_slug:
                try:
                    conflict = self.find_identifier(new_slug)
                    raise AlreadyExistsError(
                        "Canary {} already exists with identifier {}".
                        format(new_slug, conflict))
                except KeyError:
                    pass
                updates['slug'] = new_slug
            updates['name'] = name

        if periodicity is not None and periodicity != canary['periodicity']:
            if not isinstance(periodicity, Number):
                raise TypeError('periodicity must be a number')
            if periodicity <= 0:
                raise TypeError('periodicity must be positive')
            updates['periodicity'] = periodicity

            if not canary['paused']:
                updates['deadline'] = canary['history'][0][0] + \
                    datetime.timedelta(seconds=periodicity)
                is_late = updates['deadline'] < datetime.datetime.utcnow()
                if is_late != canary['late']:
                    updates['late'] = is_late
                    notify = True

        if description is not None and description != canary['description']:
            if not isinstance(description, str):
                raise TypeError('description must be a string')
            updates['description'] = description

        if emails is None:  # dummy caller specified None
            emails = []
        if emails is not None and set(emails) != set(canary['emails']):
            if isinstance(emails, str):
                raise TypeError('emails should be a list of zero or more '
                                'addresses')
            updates['emails'] = list(emails)

        if not updates:
            raise ValueError('No updates specified')

        self.store.update(identifier, updates)
        canary.update(updates)

        log.info('Updated canary {} ({}, {})'.format(
            canary['name'], identifier, canary_log_string(updates)))

        if notify:
            self.notify(canary)

        self.schedule_next_deadline()

        return canary

    def trigger(self, identifier, comment=None):
        canary = self.store.get(identifier)
        updates = {}
        was_late = canary['late']
        was_paused = canary['paused']

        if comment:
            comment = 'Triggered ({})'.format(comment)
        else:
            comment = 'Triggered'

        history = copy(canary['history'])
        self.add_history(history, comment)
        updates['history'] = history

        updates['deadline'] = history[0][0] + \
            datetime.timedelta(seconds=canary['periodicity'])
        if canary['late']:
            updates['late'] = False
        if canary['paused']:
            updates['paused'] = False

        self.store.update(identifier, updates)
        canary.update(updates)

        log.info('Triggered canary {} ({}, {}, {})'.format(
            canary['name'], identifier, comment, canary_log_string(updates)))

        if 'late' in updates:
            self.notify(canary)

        self.schedule_next_deadline()

        return (was_late, was_paused)

    def pause(self, identifier, comment=None):
        canary = self.store.get(identifier)
        updates = {}

        if canary['paused']:
            raise AlreadyPausedError()

        updates['paused'] = True

        if comment:
            comment = 'Paused ({})'.format(comment)
        else:
            comment = 'Paused'

        history = copy(canary['history'])
        self.add_history(history, comment)
        updates['history'] = history

        updates['deadline'] = None

        if canary['late']:
            updates['late'] = False

        self.store.update(identifier, updates)
        canary.update(updates)

        log.info('Paused canary {} ({}, {}, {})'.format(
            canary['name'], identifier, comment, canary_log_string(updates)))

        self.schedule_next_deadline()

        return canary

    def unpause(self, identifier, comment=None):
        canary = self.store.get(identifier)
        updates = {}

        if not canary['paused']:
            raise AlreadyUnpausedError()

        updates['paused'] = False

        if comment:
            comment = 'Unpaused ({})'.format(comment)
        else:
            comment = 'Unpaused'

        history = copy(canary['history'])
        self.add_history(history, comment)
        updates['history'] = history

        updates['deadline'] = history[0][0] + \
            datetime.timedelta(seconds=canary['periodicity'])

        self.store.update(identifier, updates)
        canary.update(updates)

        log.info('Unpaused canary {} ({}, {}, {})'.format(
            canary['name'], identifier, comment, canary_log_string(updates)))

        self.schedule_next_deadline()

        return canary

    def delete(self, identifier):
        canary = self.store.get(identifier)
        self.store.delete(identifier)

        log.info('Deleted canary {} ({})'.format(canary['name'], identifier))

        self.schedule_next_deadline()

    def get(self, identifier):
        return self.store.get(identifier)

    def list(self, *, verbose=False, paused=None, late=None, search=None):
        """N.B.: Returns an iterator."""
        return self.store.list(
            verbose=verbose,
            paused=paused,
            late=late,
            search=search,
        )

    def notify(self, canary):
        if canary['late']:
            subject = '[LATE] {} has not reported'.format(canary['name'])
        else:
            subject = '[RESUMED] {} is reporting again'.format(canary['name'])

        if not canary['emails']:
            log.info('No emails for canary {} ({}, {})'.format(
                canary['name'], canary['id'], subject))
            return

        body = ''

        if canary['late']:
            body += 'The canary {} ({}) was expected to report before {}.\n'.\
                format(canary['name'], canary['id'], canary['deadline'])
        else:
            body += 'The canary {} ({}) is reporting again as of {}.\n'.\
                format(canary['name'], canary['id'],
                       canary['history'][0][0])
            body += '\nThe next trigger for this canary is due before {}.\n'.\
                format(canary['deadline'])

        body += '\nRecent events for this canary:\n\n'

        for event in canary['history'][0:15]:
            # For some reason, when I omit the str() wrapper around
            # the datetime, the resulting string contains "30" instead
            # of the stringified datetime. I'm sure there's a good
            # reason for this, but I can't figure out what it is.
            body += '{:30} {}\n'.format(str(event[0]), event[1])

        try:
            smtp = smtplib.SMTP()
            smtp.connect()
            smtp.sendmail(self.email_sender, canary['emails'],
                          'From: {}\nTo: {}\nSubject: {}\n\n{}'.format(
                              self.email_sender,
                              ', '.join(canary['emails']),
                              subject,
                              body))
            smtp.quit()
        except:
            log.exception('Notify failed for canary {} ({}, {})'.format(
                canary['name'], canary['id'], subject))
        else:
            log.info('Notified for canary {} ({}, {})'.format(
                canary['name'], canary['id'], subject))

    def schedule_next_deadline(self, canary=None):
        if not canary:
            try:
                canary = next(self.store.upcoming_deadlines())
            except StopIteration:
                return
        when = max(1, (canary['deadline'] - datetime.datetime.utcnow()).
                   total_seconds())

        log.info('Setting alarm for canary {} ({}) at {}'.format(
            canary['name'], canary['id'], str(canary['deadline'])))
        signal.signal(signal.SIGALRM, self.deadline_handler)
        signal.alarm(int(math.ceil(when)))

    def deadline_handler(self, signum, frame):
        now = datetime.datetime.utcnow()

        for canary in self.store.upcoming_deadlines():
            if canary['deadline'] <= now:
                updates = {'late': True}
                self.store.update(canary['id'], updates)
                canary.update(updates)
                self.notify(canary)
            else:
                self.schedule_next_deadline(canary)
                return

    def slug(self, name):
        name = name.lower()
        name = re.sub(r'[-\s_]+', '-', name)
        name = re.sub(r'[^-\w]+', '', name)
        return name

    def find_identifier(self, name=None, slug=None, identifier=None):
        num_specified = sum(1 for x in (name, slug, identifier) if x)
        if not num_specified:
            raise Exception("Must specify name, slug, or identifier")
        if num_specified > 1:
            raise Exception("Specify only one of name, slug, or identifier")

        if identifier:
            return identifier

        if name:
            slug = self.slug(name)

        return self.store.find_identifier(slug)

    def create_identifier(self):
        while True:
            identifier = ''.join((choice('abcdefghijklmnopqrstuvwxyz')
                                  for c in range(8)))
            try:
                self.store.get(identifier)
            except KeyError:
                return identifier

    def add_history(self, history, comment):
        if comment is None:  # Be generous in what you accept...
            comment = ''
        if not isinstance(comment, str):
            raise TypeError('comment must be a string')

        now = datetime.datetime.utcnow()
        one_week_ago = now - datetime.timedelta(days=7)

        history.insert(0, (now, comment))

        while len(history) > 1000 or (len(history) > 100 and
                                      history[-1][0] < one_week_ago):
            history.pop()


def canary_log_string(canary):
    new_canary = copy(canary)
    if 'history' in canary and canary['history']:
        new_canary['history'] = [(str(canary['history'][0][0]),
                                  canary['history'][0][1])]
        if len(canary['history']) > 1:
            new_canary['history'].append('...')
    return str(new_canary)



#!/usr/bin/env python

# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

"""
Coal Mine CLI
"""

import argparse
from configparser import SafeConfigParser
import os
import pprint
import requests
import sys

config_file = '~/.coal-mine.ini'
config_section = 'coal-mine'


def main():
    config = SafeConfigParser()
    config.read([os.path.expanduser(config_file)])
    try:
        section = config[config_section]
    except KeyError:
        config['coal-mine'] = {}
        section = config['coal-mine']

    connect_parser = argparse.ArgumentParser(add_help=False)
    connect_parser.add_argument('--host', action='store',
                                help="Server host name",
                                default=section.get('host', 'localhost'))
    connect_parser.add_argument('--port', action='store', type=int,
                                help='Server port',
                                default=section.getint('port', 80))
    auth_key_group = connect_parser.add_mutually_exclusive_group()
    auth_key_group.add_argument('--auth-key', action='store',
                                help='Authentication key',
                                default=section.get('auth-key', None))
    auth_key_group.add_argument('--no-auth-key', action='store_true',
                                help='Disable authentication',
                                default=False)

    parser = argparse.ArgumentParser(description="CLI wrapper for Coal "
                                     "Mine's HTTP API")
    subparsers = parser.add_subparsers()

    configure_parser = subparsers.add_parser('configure', help='Save '
                                             'configuration from command line '
                                             'to ' + config_file,
                                             parents=[connect_parser])
    configure_parser.set_defaults(func=handle_configure,
                                  config_parser=config)

    create_parser = subparsers.add_parser('create', help='Create canary',
                                          parents=[connect_parser])
    create_parser.add_argument('--name', action='store', required=True)
    create_parser.add_argument('--periodicity', action='store', type=int,
                               required=True)
    create_parser.add_argument('--description', action='store')
    create_parser.add_argument('--email', action='append')
    create_parser.add_argument('--paused', action='store_true', default=False)
    create_parser.set_defaults(func=handle_create)

    id_parser = argparse.ArgumentParser(add_help=False)
    id_parser_group = id_parser.add_mutually_exclusive_group(required=True)
    id_parser_group.add_argument('--name', action='store')
    id_parser_group.add_argument('--slug', action='store')
    id_parser_group.add_argument('--id', action='store')

    delete_parser = subparsers.add_parser('delete', help='Delete canary',
                                          parents=[connect_parser, id_parser])
    delete_parser.set_defaults(func=handle_delete)

    update_parser = subparsers.add_parser('update', help='Update canary',
                                          parents=[connect_parser])
    update_parser.add_argument('--name', action='store')
    update_parser_group = update_parser.add_mutually_exclusive_group()
    update_parser_group.add_argument('--slug', action='store')
    update_parser_group.add_argument('--id', action='store')
    update_parser.add_argument('--periodicity', action='store', type=int)
    update_parser.add_argument('--description', action='store')
    update_parser.add_argument('--email', action='append')
    update_parser.set_defaults(func=handle_update)

    get_parser = subparsers.add_parser('get', help='Get canary',
                                       parents=[connect_parser, id_parser])
    get_parser.set_defaults(func=handle_get)

    list_parser = subparsers.add_parser('list', help='List canaries',
                                        parents=[connect_parser])
    list_parser.add_argument('--verbose', action='store_true', default=None)
    paused_group = list_parser.add_mutually_exclusive_group()
    paused_group.add_argument('--paused', action='store_true', default=None)
    paused_group.add_argument('--no-paused', dest='paused',
                              action='store_false', default=None)
    late_group = list_parser.add_mutually_exclusive_group()
    late_group.add_argument('--late', action='store_true', default=None)
    late_group.add_argument('--no-late', dest='late',
                            action='store_false', default=None)
    list_parser.add_argument('--search', action='store', default=None,
                             help='Regular expression to match against name, '
                             'slug, and identifier')
    list_parser.set_defaults(func=handle_list)

    trigger_parser = subparsers.add_parser('trigger', help='Trigger canary',
                                           parents=[connect_parser, id_parser])
    trigger_parser.add_argument('--comment', action='store')
    trigger_parser.set_defaults(func=handle_trigger)

    pause_parser = subparsers.add_parser('pause', help='Pause canary',
                                         parents=[connect_parser, id_parser])
    pause_parser.add_argument('--comment', action='store')
    pause_parser.set_defaults(func=handle_pause)

    unpause_parser = subparsers.add_parser('unpause', help='Unpause canary',
                                           parents=[connect_parser, id_parser])
    unpause_parser.add_argument('--comment', action='store')
    unpause_parser.set_defaults(func=handle_unpause)

    args = parser.parse_args()
    if args.no_auth_key:
        args.auth_key = None
    del args.no_auth_key
    args.func(args)


def handle_configure(args):
    section = args.config_parser[config_section]
    section['host'] = args.host
    section['port'] = str(args.port)
    if args.auth_key:
        section['auth-key'] = args.auth_key
    elif args.no_auth_key:
        section.pop('auth-key', None)
    with open(os.path.expanduser(config_file), 'w') as configfile:
        args.config_parser.write(configfile)


def handle_create(args):
    call('create', args)


def handle_delete(args):
    call('delete', args)


def handle_update(args):
    if args.id and args.slug:
        sys.exit("Don't specify both --id and --slug")
    elif not (args.name or args.id or args.slug):
        sys.exit('Must specify --name, --id, or --slug')
    if args.name and not (args.id or args.slug):
        found = call('get', args, {'name': args.name}, action='return')
        del args.name
        args.id = found['canary']['id']
    call('update', args)


def handle_get(args):
    call('get', args)


def handle_list(args):
    if args.paused is None:
        del args.paused
    if args.late is None:
        del args.late
    if args.search is None:
        del args.search
    call('list', args)


def handle_trigger(args):
    del args.auth_key
    call('trigger', args)


def handle_pause(args):
    call('pause', args)
    pass


def handle_unpause(args):
    call('unpause', args)
    pass


def call(command, args, payload=None, action='print'):
    url = 'http://{}:{}/coal-mine/v1/canary/{}'.format(
        args.host, args.port, command)
    if payload:
        if args.auth_key:
            payload['auth_key'] = args.auth_key
    else:
        payload = {key: (getattr(args, key) if key == 'email'
                         else str(getattr(args, key)))
                   for key in dir(args)
                   if key not in ('host', 'port', 'func') and
                   not key.startswith('_') and getattr(args, key) is not None}
    response = requests.get(url, params=payload)
    if response.status_code != 200:
        sys.stderr.write('{} {}\n'.format(
            response.status_code, response.reason))
        try:
            pprint.pprint(response.json(), sys.stderr)
        except:
            sys.stderr.write(response.text + '\n')
        sys.exit(1)
    if action == 'print':
        pprint.pprint(response.json())
    elif action == 'return':
        return response.json()
    else:
        raise Exception('Unrecognized action: {}'.format(action))

if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

"""
Coal Mine WSGI server
"""

from coal_mine.business_logic import BusinessLogic
from cgi import parse_qs
from configparser import SafeConfigParser, NoSectionError, NoOptionError
from functools import wraps
import json
import logbook
from coal_mine.mongo_store import MongoStore
import os
import re
import sys
from wsgiref.simple_server import make_server, WSGIRequestHandler

config_file = 'coal-mine.ini'
url_prefix = '/coal-mine/v1/canary/'

log = logbook.Logger('coal-mine')
business_logic = None
auth_key = None


def main():
    global business_logic, auth_key

    config = SafeConfigParser()
    dirs = ('.', '/etc', '/usr/local/etc')
    if not config.read([os.path.join(dir, config_file) for dir in dirs]):
        sys.exit('Could not find {} in {}'.format(config_file, dirs))

    try:
        logfile = config.get('logging', 'file')
        rotating = config.getboolean('logging', 'rotate', fallback=False)
        if rotating:
            max_size = config.get('logging', 'max_size', fallback=1048576)
            backup_count = config.get('logging', 'backup_count', fallback=5)
            handler = logbook.RotatingFileHandler(logfile, max_size=max_size,
                                                  backup_count=backup_count)
        else:
            handler = logbook.FileHandler(logfile)
        handler.push_application()
    except:
        pass

    try:
        kwargs = dict(config.items('mongodb'))
    except NoSectionError:
        sys.exit('No "mongodb" section in config file')
    args = []
    for arg in ('hosts', 'database', 'username', 'password'):
        try:
            args.append(config.get('mongodb', arg))
        except NoOptionError:
            sys.exit('No "{}" setting in "mongodb" section of config file'.
                     format(arg))
        kwargs.pop(arg)
    store = MongoStore(*args, **kwargs)

    try:
        email_sender = config.get('email', 'sender')
    except NoSectionError:
        sys.exit('No "email" section in config file')
    except NoOptionError:
        sys.exit('No "sender" setting in "email" section of config file')

    business_logic = BusinessLogic(store, email_sender)

    try:
        listen_port = int(config.get('wsgi', 'port'))
        log.info('Binding to port {}'.format(listen_port))
    except:
        listen_port = 80
        log.info('Binding to default port {}'.format(listen_port))

    try:
        auth_key = config.get('wsgi', 'auth_key')
        log.info('Server authentication enabled')
    except:
        log.warning('Server authentication DISABLED')

    httpd = make_server('', listen_port, application,
                        handler_class=LogbookWSGIRequestHandler)
    business_logic.schedule_next_deadline()
    httpd.serve_forever()


def application(environ, start_response):
    handlers = {
        'create': handle_create,
        'delete': handle_delete,
        'update': handle_update,
        'get': handle_get,
        'list': handle_list,
        'trigger': handle_trigger,
        'pause': handle_pause,
        'unpause': handle_unpause,
    }

    path_info = environ['PATH_INFO']
    # Special case: make trigger URLs easy
    match = re.match(r'/([a-z]{8})$', path_info)
    if match:
        id = match.group(1)
        qs = 'id={}'.format(id)
        if environ['QUERY_STRING']:
            environ['QUERY_STRING'] += '&' + qs
        else:
            environ['QUERY_STRING'] = qs
        path_info = environ['PATH_INFO'] = url_prefix + 'trigger'

    if not path_info.startswith(url_prefix):
        start_response('404 Not Found', headers=[])
        return []

    command = path_info[len(url_prefix):]

    if command not in handlers:
        start_response('404 Not Found', headers=[])
        return []

    q = parse_qs(environ['QUERY_STRING'])
    if auth_key and command != 'trigger' and \
       q.pop('auth_key', [None])[0] != auth_key:
        start_response('401 Unauthorized', headers=[])
        return []

    (status_code, data) = handlers[command](q, start_response)

    data = [json.dumps(data, indent=4).encode('utf-8'),
            '\n'.encode('utf-8')]

    start_response(status_code,
                   headers=[('Content-Type', 'text/json; charset=utf-8')])
    return data


def required_parameters(*args):
    def decorator(f):
        @wraps(f)
        def wrapper(query, start_response):
            for arg in args:
                if arg not in query:
                    raise Exception('Missing argument "{}"'.format(arg))
            return f(query, start_response)
        return wrapper
    return decorator


def find_identifier(name_ok=True):
    def decorator(f):
        @wraps(f)
        def wrapper(query, start_response):
            name = slug = identifier = None
            if 'id' in query:
                identifier = query.pop('id')[-1]
            elif 'slug' in query:
                slug = query.pop('slug')[-1]
            elif name_ok and 'name' in query:
                name = query.pop('name')[-1]
            if not (name or slug or identifier):
                if name_ok:
                    raise Exception('Must specify id, slug, or name')
                else:
                    raise Exception('Muset specify id or slug')
            query['id'] = business_logic.find_identifier(
                name, slug, identifier)
            return f(query, start_response)
        return wrapper
    return decorator


def string_parameters(*args):
    def decorator(f):
        @wraps(f)
        def wrapper(query, start_response):
            for arg in args:
                if arg in query:
                    query[arg] = query[arg][-1]
            return f(query, start_response)
        return wrapper
    return decorator


def int_parameters(*args):
    def decorator(f):
        @wraps(f)
        def wrapper(query, start_response):
            for arg in args:
                if arg in query:
                    query[arg] = int(query[arg][-1])
            return f(query, start_response)
        return wrapper
    return decorator


def boolean_parameters(*args):
    def decorator(f):
        @wraps(f)
        def wrapper(query, start_response):
            for arg in args:
                if arg not in query:
                    continue
                val = query[arg][-1]
                if val.lower() in ('true', 'yes', '1'):
                    query[arg] = True
                elif val.lower() in ('false', 'no', '0', ''):
                    query[arg] = False
                else:
                    raise Exception(
                        'Bad boolean value "{}" for parameter "{}"'.format(
                            val, arg))
            return f(query, start_response)
        return wrapper
    return decorator


def valid_parameters(*args):
    def decorator(f):
        @wraps(f)
        def wrapper(query, start_response):
            for arg in query:
                if arg not in args:
                    raise Exception('Unexpected argument "{}"'.format(arg))
            return f(query, start_response)
        return wrapper
    return decorator


def handle_exceptions(f):
    @wraps(f)
    def wrapper(query, start_response):
        try:
            return f(query, start_response)
        except Exception as e:
            log.exception('Exception in {}'.format(f))
            return ('400 Bad Request', {'status': 'error', 'error': repr(e)})
    return wrapper


@handle_exceptions
@required_parameters('name', 'periodicity')
@string_parameters('name', 'description')
@int_parameters('periodicity')
@boolean_parameters('paused')
@valid_parameters('name', 'periodicity', 'description', 'email', 'paused')
def handle_create(query, start_response):
    canary = business_logic.create(query['name'],
                                   query['periodicity'],
                                   query.get('description', ''),
                                   query.get('email', []),
                                   query.get('paused', False))
    return ('200 OK', {'status': 'ok', 'canary': jsonify_canary(canary)})


@handle_exceptions
@find_identifier()
@valid_parameters('id')
def handle_delete(query, start_response):
    business_logic.delete(query['id'])
    return ('200 OK', {'status': 'ok'})


@handle_exceptions
@find_identifier(name_ok=False)
@string_parameters('name', 'description')
@int_parameters('periodicity')
@valid_parameters('id', 'name', 'periodicity', 'description', 'email')
def handle_update(query, start_response):
    canary = business_logic.update(query['id'],
                                   query.get('name', None),
                                   query.get('periodicity', None),
                                   query.get('description', None),
                                   query.get('email', None))
    return ('200 OK', {'status': 'ok', 'canary': jsonify_canary(canary)})


@handle_exceptions
@find_identifier()
@valid_parameters('id')
def handle_get(query, start_response):
    canary = business_logic.get(query['id'])

    return ('200 OK', {'status': 'ok', 'canary': jsonify_canary(canary)})


@handle_exceptions
@boolean_parameters('verbose', 'paused', 'late')
@string_parameters('search')
@valid_parameters('verbose', 'paused', 'late', 'search')
def handle_list(query, start_response):
    canaries = [jsonify_canary(canary)
                for canary in business_logic.list(
                    verbose=query.get('verbose', False),
                    paused=query.get('paused', None),
                    late=query.get('late', None),
                    search=query.get('search', None))]
    return ('200 OK', {'status': 'ok', 'canaries': canaries})


@handle_exceptions
@find_identifier()
@string_parameters('comment', 'm')
@valid_parameters('id', 'comment', 'm')
def handle_trigger(query, start_response):
    comment = query.get('comment', query.get('m', ''))
    (recovered, unpaused) = business_logic.trigger(query['id'], comment)
    return ('200 OK',
            {'status': 'ok', 'recovered': recovered, 'unpaused': unpaused})


@handle_exceptions
@find_identifier()
@string_parameters('comment')
@valid_parameters('id', 'comment')
def handle_pause(query, start_response):
    canary = business_logic.pause(query['id'], query.get('comment', ''))
    return ('200 OK', {'status': 'ok', 'canary': jsonify_canary(canary)})


@handle_exceptions
@find_identifier()
@string_parameters('comment')
@valid_parameters('id', 'comment')
def handle_unpause(query, start_response):
    canary = business_logic.unpause(query['id'], query.get('comment', ''))
    return ('200 OK', {'status': 'ok', 'canary': jsonify_canary(canary)})


def jsonify_canary(canary):
    for key, value in [(k, v) for k, v in canary.items()]:
        if value is None:
            del canary[key]

    if 'deadline' in canary:
        canary['deadline'] = canary['deadline'].isoformat()

    if 'history' in canary:
        canary['history'] = tuple((d.isoformat(), c)
                                  for d, c in canary['history'])
    return canary


class LogbookWSGIRequestHandler(WSGIRequestHandler):
    def log_message(self, format, *args):
        msg = format % args
        msg = re.sub(r'\b(auth_key=)[^&;]+', r'\1<key>', msg)
        log.info("%s - - %s" % (self.address_string(), msg))


if __name__ == '__main__':
    main()



# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

"""
MongoDB store for Coal Mine
"""

from coal_mine.abstract_store import AbstractStore
import bson
from copy import copy
from logbook import Logger
import pymongo
import re
import time

log = Logger('MongoStore')


class MongoStore(AbstractStore):
    def __init__(self, hosts, database, username, password, **kwargs):
        """Keyword arguments are the same as what would be passed to
        MongoClient or MongoReplicaSetClient."""

        if 'replicaSet' in kwargs:
            client = pymongo.MongoReplicaSetClient
        else:
            client = pymongo.MongoClient
        connection = client(hosts, **kwargs)
        db = connection[database]
        if username or password:
            db.authenticate(username, password)
        self.db = db
        self.collection = self.db['canaries']

        self.collection.ensure_index([('id', pymongo.ASCENDING)],
                                     unique=True)

        # for list()
        self.collection.ensure_index([('paused', pymongo.ASCENDING),
                                      ('late', pymongo.ASCENDING),
                                      ('deadline', pymongo.ASCENDING)])
        self.collection.ensure_index([('paused', pymongo.ASCENDING),
                                      ('deadline', pymongo.ASCENDING)])
        self.collection.ensure_index([('late', pymongo.ASCENDING),
                                      ('deadline', pymongo.ASCENDING)])

        # for find_identifier(), as well as to ensure unique slugs
        self.collection.ensure_index([('slug', pymongo.ASCENDING)],
                                     unique=True)

    def create(self, canary):
        canary['_id'] = bson.ObjectId()
        while True:
            try:
                self.collection.save(canary)
                del canary['_id']
                break
            except pymongo.errors.AutoReconnect:
                log.exception('save failure, retrying')
                time.sleep(1)

    def update(self, identifier, updates):
        updates = copy(updates)
        unset = {}
        for key, value in [(k, v) for k, v in updates.items()]:
            if value is None:
                del updates[key]
                unset[key] = ''
        doc = {'$set': updates}
        if unset:
            doc['$unset'] = unset
        while True:
            try:
                self.collection.update({'id': identifier}, doc)
                return
            except pymongo.errors.AutoReconnect:
                log.exception('update failure, retrying')
                time.sleep(1)

    def get(self, identifier):
        while True:
            try:
                canary = self.collection.find_one({'id': identifier},
                                                  fields={'_id': False})
                if not canary:
                    raise KeyError('No such canary {}'.format(identifier))
                return canary
            except pymongo.errors.AutoReconnect:
                log.exception('find_one failure, retrying')
                time.sleep(1)

    def list(self, *, verbose=False, paused=None, late=None, search=None,
             order_by=None):
        if verbose:
            fields = {'_id': False}
        else:
            fields = {'_id': False, 'name': True, 'id': True}

        spec = {}

        if paused is not None:
            spec['paused'] = paused

        if late is not None:
            spec['late'] = late

        if order_by is not None:
            order_by = [(order_by, pymongo.ASCENDING)]

        if search is not None:
            search = re.compile(search)
            spec['$or'] = [{'name': search}, {'slug': search}, {'id': search}]

        skip = 0

        while True:
            try:
                for canary in self.collection.find(spec, fields=fields,
                                                   sort=order_by, skip=skip):
                    yield canary
                break
            except pymongo.errors.AutoReconnect:
                log.exception('find failure, retrying')
                time.sleep(1)

    def upcoming_deadlines(self):
        return self.list(verbose=True, paused=False, late=False,
                         order_by='deadline')

    def delete(self, identifier):
        while True:
            try:
                result = self.collection.remove({'id': identifier})
                if result['n'] == 0:
                    raise KeyError('No such canary {}'.format(identifier))
                return
            except pymongo.errors.AutoReconnect:
                log.exception('remove failure, retrying')
                time.sleep(1)

    def find_identifier(self, slug):
        while True:
            try:
                result = self.collection.find_one({'slug': slug})
                if not result:
                    raise KeyError('No such canary {}'.format(slug))
                return result['id']
            except pymongo.errors.AutoReconnect:
                log.exception('find_one failure, retrying')
                time.sleep(1)



# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

"""
Abstract store for Coal Mine

Subclass for a specific storage engine.
"""

from abc import ABCMeta, abstractmethod


class AlreadyExistsError(Exception):
    pass


class AbstractStore(object, metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        """Args and behavior are dependent on the storage engine."""
        raise NotImplementedError('__init__')

    @abstractmethod
    def create(self, canary):
        """Should return string identifier of created canary."""
        raise NotImplementedError('create')

    @abstractmethod
    def update(self, identifier, updates):
        raise NotImplementedError('update')

    @abstractmethod
    def get(self, identifier):
        """Should raise KeyError if not found, or return a dict with these
        keys: id, name, description, slug, periodicity, emails, late,
        paused, deadline, history. History should be a list of tuples,
        each of which contains a naive UTC timestamp and a possibly
        empty comment, sorted from most to least recent. Deadline
        should be a naive UTC timestamp."""
        raise NotImplementedError('get')

    @abstractmethod
    def list(self, *, verbose=False, paused=None, late=None, search=None):
        """Return an iterator which yields dicts. If verbose is False, then
        the dicts contain only name and id, otherwise, all fields
        (same as returned by get()) are returned. If paused, late,
        and/or search are specified, they are used to filter the
        results. The latter is a regular expression (string, not
        regular expression object), which is matched against the name,
        slug, and id of canaries and only matches are returned."""
        raise NotImplementedError('list')

    @abstractmethod
    def upcoming_deadlines(self):
        """Return an iterator which yields canaries (same as returned by
        get()) that are unpaused and not yet late, sorted by deadline
        in increasing order, i.e., the canary that will pass its
        deadline soonest is returned first."""
        raise NotImplementedError('upcoming_deadlines')

    @abstractmethod
    def delete(self, identifier):
        """Raise KeyError if a canary with the specified identifier
        doesn't exist."""
        raise NotImplementedError('delete')

    @abstractmethod
    def find_identifier(self, slug):
        """Should raise KeyError if a canary with the specified slug
        does not exist, or return the identifier string."""
        raise NotImplementedError('find_identifier')



import sys
from pypandoc import convert
from setuptools import setup, find_packages

if sys.version_info[0] < 3:
    sys.exit("This package does not work with Python 2.")

setup(
    name="coal_mine",
    version='0.2.2',
    author='Quantopian Inc.',
    author_email='opensource@quantopian.com',
    description="Coal Mine - Periodic task execution monitor",
    url='https://github.com/quantopian/coal-mine',
    long_description=convert('README.md', 'rst'),
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Server',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Systems Administration',
    ],
    packages=find_packages(),
    setup_requires=open('requirements_dev.txt').read(),
    install_requires=open('requirements.txt').read(),
    entry_points={
        'console_scripts': [
            "coal-mine = coal_mine.server:main",
            "cmcli = coal_mine.cli:main"
        ]
    },
    zip_safe=True,
)
